open Context
open Structure

(* Transform a tree by removing redundant tags (bold if already in bold, etc) *)
let transform tree_in =
	let is_color_only subseq =
		(* if [subseq] is composed of only a color change, then true
		 * else false
		 *)
		match subseq with
		| [ Red s ] -> true
		| [ Green s ] -> true
		| [ Blue s ] -> true
		| [ Cyan s ] -> true
		| [ Magenta s ] -> true
		| [ Yellow s ] -> true
		| [ Black s ] -> true
		| [ White s ] -> true
		| _ -> false
	in
	let rec transform_r tr context value =
		match tr with
		| [] ->
			List.rev value
		| hd :: rest ->
			match hd with
			| Text st ->
				transform_r rest context (hd :: value)
			| Bold seq ->
				if context.b then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with b=true} [] in
					transform_r rest context ((Bold newhd) :: value)
				end
			| Emphasis seq ->
				let ctx =
					if context.s then context
					else {context with em=not context.em}
				in
				let newhd = transform_r seq ctx [] in
				transform_r rest context ((Emphasis newhd) :: value)
			| Italic seq ->
				if context.i then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with i=true} [] in
					transform_r rest context ((Italic newhd) :: value)
				end
			| Plain seq ->
				if not context.s && not context.em && not context.i
						&& not context.b && not context.tt && context.u = 0
						then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq
							{context with s=false; em=false;
							 i=false; b=false;
							 tt=false; u=0} [] in
					transform_r rest context ((Plain newhd) :: value)
				end
			| Strong seq ->
				if context.s && not context.em then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq
							{context with s=true; em=false} [] in
					transform_r rest context ((Strong newhd) :: value)
				end
			| Typewriter seq ->
				if context.tt then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with tt=true} [] in
					transform_r rest context ((Typewriter newhd) :: value)
				end
			| Underline seq ->
				if context.u < 3 then begin
					let ctx = {context with u=(context.u + 1)} in
					let newhd = transform_r seq ctx [] in
					transform_r rest context ((Underline newhd) :: value)
				end else begin
					transform_r (seq @ rest) context value
				end
			| Size (sz,seq) ->
				if context.size = sz then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with size=sz} [] in
					transform_r rest context ((Size (sz,newhd)) :: value)
				end
			| Red seq ->
				if context.color = R then
					transform_r (seq @ rest) context value
				else if (is_color_only seq) then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with color=R} [] in
					transform_r rest context ((Red newhd) :: value)
				end
			| Green seq ->
				if context.color = G then
					transform_r (seq @ rest) context value
				else if (is_color_only seq) then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with color=G} [] in
					transform_r rest context ((Green newhd) :: value)
				end
			| Blue seq ->
				if context.color = B then
					transform_r (seq @ rest) context value
				else if (is_color_only seq) then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with color=B} [] in
					transform_r rest context ((Blue newhd) :: value)
				end
			| Cyan seq ->
				if context.color = C then
					transform_r (seq @ rest) context value
				else if (is_color_only seq) then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with color=C} [] in
					transform_r rest context ((Cyan newhd) :: value)
				end
			| Magenta seq ->
				if context.color = M then
					transform_r (seq @ rest) context value
				else if (is_color_only seq) then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with color=M} [] in
					transform_r rest context ((Magenta newhd) :: value)
				end
			| Yellow seq ->
				if context.color = Y then
					transform_r (seq @ rest) context value
				else if (is_color_only seq) then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with color=Y} [] in
					transform_r rest context ((Yellow newhd) :: value)
				end
			| Black seq ->
				if context.color = K then
					transform_r (seq @ rest) context value
				else if (is_color_only seq) then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with color=K} [] in
					transform_r rest context ((Black newhd) :: value)
				end
			| White seq ->
				if context.color = W then
					transform_r (seq @ rest) context value
				else if (is_color_only seq) then
					transform_r (seq @ rest) context value
				else begin
					let newhd = transform_r seq {context with color=W} [] in
					transform_r rest context ((White newhd) :: value)
				end
	in
	let result = transform_r tree_in
			{b=false; em=false; i=false; s=false; tt=false;
			u=0; size=0; color=Any} [] in
	result
