open Context
open Structure

let _PAT_EWS = Str.regexp "[ \009\013\010]$"
let _PAT_BWS = Str.regexp "^[ \009\013\010]"

(* Transform a tree by removing one of the spaces surrounding an end
 * tag if it's a resonable thing to do
 *)
let transform tree_in =
	let do_change seq1 ctx seq2 =
		try
			(* If the last segment of seq1 is a TEXT and the
			 * first segment of seq2 is a TEXT
			 * and the first TEXT ends with a space and the second
			 * TEXT begins with a space, then we have a candidate.
			 *)
			match (List.rev seq1) with
			| Text t1 :: rseq1 ->
				begin
				match seq2 with
				| Text t2 :: rseq2 ->
					ignore(Str.search_forward _PAT_EWS t1 0);
					ignore(Str.search_forward _PAT_BWS t2 0);
					(* Okay ... we have a candidate pair. We can eliminate
					 * the ending space in [t1] if the context allows it
					 *)
					if ctx.tt then raise Not_found;
					let nt1 = Str.global_replace _PAT_EWS "" t1 in
					(List.rev rseq1) @ [Text nt1]
				| _ ->
					raise Not_found
				end
			| _ ->
				raise Not_found
		with Not_found ->
			seq1
	in
	let rec transform_r tr context value =
		match tr with
		| [] ->
			List.rev value
		| hd :: rest ->
			match hd with
			| Text st ->
				transform_r rest context (hd :: value)
			| Bold seq ->
				let ctx = {context with b=true} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Bold newhd) :: value)
			| Emphasis seq ->
				let ctx =
					if context.s then context
					else {context with em=not context.em}
				in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Emphasis newhd) :: value)
			| Italic seq ->
				let ctx = {context with i=true} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Italic newhd) :: value)
			| Plain seq ->
				let ctx = {context with s=false; em=false;
						 i=false; b=false;
						 tt=false; u=0} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Plain newhd) :: value)
			| Strong seq ->
				let ctx = {context with s=true; em=false} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Strong newhd) :: value)
			| Typewriter seq ->
				let ctx = {context with tt=true} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Typewriter newhd) :: value)
			| Underline seq ->
				let newval =
					if context.u < 3 then (context.u + 1)
					else context.u
				in
				let ctx = {context with u=newval} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Underline newhd) :: value)
			| Size (sz,seq) ->
				let ctx = {context with size=sz} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Size (sz,newhd)) :: value)
			| Red seq ->
				let ctx = {context with color=R} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Red newhd) :: value)
			| Green seq ->
				let ctx = {context with color=G} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Green newhd) :: value)
			| Blue seq ->
				let ctx = {context with color=B} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Blue newhd) :: value)
			| Cyan seq ->
				let ctx = {context with color=C} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Cyan newhd) :: value)
			| Magenta seq ->
				let ctx = {context with color=M} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Magenta newhd) :: value)
			| Yellow seq ->
				let ctx = {context with color=Y} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Yellow newhd) :: value)
			| Black seq ->
				let ctx = {context with color=K} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((Black newhd) :: value)
			| White seq ->
				let ctx = {context with color=W} in
				let newhd = do_change (transform_r seq ctx []) ctx rest in
				transform_r rest context ((White newhd) :: value)
	in
	let result = transform_r tree_in
			{b=false; em=false; i=false; s=false; tt=false;
			u=0; size=0; color=Any} [] in
	result
