open Context
open Structure

(* Transform a tree by removing useless EMs *)
let transform tree_in =
	let rec transform_r tr context value =
		match tr with
		| [] ->
			List.rev value
		| hd :: rest ->
			match hd with
			| Text st ->
				transform_r rest context ((Text st) :: value)
			| Bold seq ->
				let newhd = transform_r seq {context with b=true} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context ((Emphasis [Bold subseq]) :: value)
				| _ ->
					transform_r rest context ((Bold newhd) :: value)
				end
			| Emphasis seq ->
				if context.s then begin
					transform_r (seq @ rest) context value
				end else begin
					(* nested EMs undo each other in some cases *)
					match seq with
					| Emphasis subseq :: more ->
						transform_r (subseq @ rest) context value
					| Text " " :: Emphasis subseq :: more ->
						transform_r ((Text " ") :: subseq @ rest) context value
					| _ ->
						let ctx = {context with em=not context.em} in
						let newhd = transform_r seq ctx [] in
						transform_r rest context ((Emphasis newhd) :: value)
				end
			| Italic seq ->
				let newhd = transform_r seq {context with i=true} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context
							((Emphasis [Italic subseq]) :: value)
				| _ ->
					transform_r rest context ((Italic newhd) :: value)
				end
			| Plain seq ->
				let newhd = transform_r seq
						{context with s=false; em=false;
						 i=false; b=false;
						 tt=false; u=0} [] in
				transform_r rest context ((Plain newhd) :: value)
			| Strong seq ->
				let newhd = transform_r seq
						{context with s=true; em=false} [] in
				transform_r rest context ((Strong newhd) :: value)
			| Typewriter seq ->
				let newhd = transform_r seq {context with tt=true} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context 
							((Emphasis [Typewriter subseq]) :: value)
				| _ ->
					transform_r rest context ((Typewriter newhd) :: value)
				end
			| Underline seq ->
				let newval =
					if context.u < 3 then (context.u + 1)
					else context.u
				in
				let newhd = transform_r seq {context with u=newval} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context 
							((Emphasis [Underline subseq]) :: value)
				| _ ->
					transform_r rest context ((Underline newhd) :: value)
				end
			| Size (sz,seq) ->
				let newhd = transform_r seq {context with size=sz} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context 
							((Emphasis [Size (sz,subseq)]) :: value)
				| _ ->
					transform_r rest context ((Size (sz,newhd)) :: value)
				end
			| Red seq ->
				let newhd = transform_r seq {context with color=R} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context 
							((Emphasis [Red subseq]) :: value)
				| _ ->
					transform_r rest context ((Red newhd) :: value)
				end
			| Green seq ->
				let newhd = transform_r seq {context with color=G} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context 
							((Emphasis [Green subseq]) :: value)
				| _ ->
					transform_r rest context ((Green newhd) :: value)
				end
			| Blue seq ->
				let newhd = transform_r seq {context with color=B} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context 
							((Emphasis [Blue subseq]) :: value)
				| _ ->
					transform_r rest context ((Blue newhd) :: value)
				end
			| Cyan seq ->
				let newhd = transform_r seq {context with color=C} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context 
							((Emphasis [Cyan subseq]) :: value)
				| _ ->
					transform_r rest context ((Cyan newhd) :: value)
				end
			| Magenta seq ->
				let newhd = transform_r seq {context with color=M} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context 
							((Emphasis [Magenta subseq]) :: value)
				| _ ->
					transform_r rest context ((Magenta newhd) :: value)
				end
			| Yellow seq ->
				let newhd = transform_r seq {context with color=Y} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context 
							((Emphasis [Yellow subseq]) :: value)
				| _ ->
					transform_r rest context ((Yellow newhd) :: value)
				end
			| Black seq ->
				let newhd = transform_r seq {context with color=K} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context 
							((Emphasis [Black subseq]) :: value)
				| _ ->
					transform_r rest context ((Black newhd) :: value)
				end
			| White seq ->
				let newhd = transform_r seq {context with color=W} [] in
				begin
				match newhd with
				| [ Emphasis subseq ] ->
					transform_r rest context 
							((Emphasis [White subseq]) :: value)
				| _ ->
					transform_r rest context ((White newhd) :: value)
				end
	in
	let result = transform_r tree_in
			{b=false; em=false; i=false; s=false; tt=false;
			u=0; size=0; color=Any} [] in
	result
